#! /usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
from typing import List, Set, Tuple, Type, Union


class IdrTorchWarning(RuntimeWarning):
    """
    Type (subtype of RuntimeWarning) of all warnings raised by idr_torch.
    You can use it to customize warning filters.
    """

    pass


class WarningFilter:
    def __init__(self):
        self.registry: Set[Tuple[str, Union[Type[str], Type[Warning]]]] = set()

    def block(self, warning: Union[str, Warning]) -> bool:
        text = str(warning)
        category = warning.__class__
        if not isinstance(warning, IdrTorchWarning):
            return False
        message = (text, category)
        if message not in self.registry:
            self.registry.add(message)
            return False
        return True

    def warn(self, warning_list: List[warnings.WarningMessage]):
        for warning in warning_list:
            if not self.block(warning.message):
                category = (
                    warning.message.__class__
                    if isinstance(warning.message, Warning)
                    else Warning
                )
                warnings.warn(
                    message=str(warning.message),
                    category=category,
                    stacklevel=3,
                )


warning_filter = WarningFilter()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
from functools import wraps
from re import sub
from typing import Optional

import torch.profiler
from packaging.version import Version


class persistent_locals(object):
    """
    Allows access to local variables of a function.
    Shamelessly stolen from
    https://stackoverflow.com/questions/9186395/python-is-there-a-way-to-get-a-local-function-variable-from-within-a-decorator
    """

    def __init__(self, func):
        self._locals = {}
        self.func = func

    def __call__(self, *args, **kwargs):
        def tracer(frame, event, arg):
            if event == "return":
                self._locals = frame.f_locals.copy()

        # tracer is activated on next call, return or exception
        sys.setprofile(tracer)
        try:
            # trace the function call
            res = self.func(*args, **kwargs)
        finally:
            # disable tracer and replace with old one
            sys.setprofile(None)
        return res

    def clear_locals(self):
        self._locals = {}

    @property
    def locals(self):
        return self._locals


if Version(torch.__version__) >= Version("1.12.0"):

    @wraps(torch.profiler.tensorboard_trace_handler)
    def tensorboard_trace_handler(
        dir_name: str, worker_name: Optional[str] = None, use_gzip: bool = False
    ):
        handler_fn = torch.profiler.tensorboard_trace_handler(
            dir_name=dir_name,
            worker_name=worker_name,
            use_gzip=use_gzip,
        )
        handler_fn = persistent_locals(handler_fn)

        @wraps(handler_fn)
        def wrapper(prof):
            handler_fn(prof)
            file_name = handler_fn._locals["file_name"]
            dir_name = handler_fn._locals["dir_name"]
            handler_fn.clear_locals()
            with open(os.path.join(dir_name, file_name), "r+") as file:
                content = file.read()
                file.seek(0)
                file.truncate()

                # Restore profiler steps and dataloader
                new_content = sub("user_annotation", "cpu_op", content)

                # Restore runtime category
                new_content = sub("cuda_runtime", "runtime", new_content)

                file.write(new_content)

        return wrapper

else:
    tensorboard_trace_handler = torch.profiler.tensorboard_trace_handler



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
from functools import wraps
from typing import Type

from ...utils import IdrTorchWarning
from .. import API
from .decorate_methods import decorate_methods


def warn(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        warnings.warn(
            message=(
                "Calling idr_torch only makes sense within a distributed execution "
                "but none was detected. You may have forgotten to use a launcher "
                "(such as srun). Back to default (non distributed) values"
            ),
            category=IdrTorchWarning,
            stacklevel=4,
        )
        return func(self, *args, **kwargs)

    return wrapper


def UndistributedWarning(cls: Type[API]) -> Type[API]:
    return decorate_methods(cls, func_to_apply=warn)



from .base import API
from .default import DefaultAPI
from .modifiers import AutoMasterAddressPort, UndistributedWarning, decorate_methods
from .slurm import SlurmAPI
from .torchelastic import TorchElasticAPI

__all__ = [
    "API",
    "SlurmAPI",
    "DefaultAPI",
    "TorchElasticAPI",
    "AutoMasterAddressPort",
    "UndistributedWarning",
    "decorate_methods",
]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
from abc import ABC, abstractmethod
from typing import List, Union, TYPE_CHECKING

from ..utils import IdrTorchWarning

if TYPE_CHECKING:
    import torch


def keep_as_func(func: callable) -> callable:
    setattr(func, "__keep_as_func__", True)
    return func


class API(ABC):
    priority: int = 5000
    name: str = "AbstractAPI"

    @abstractmethod
    def is_launcher(self) -> bool:
        """
        Detects if the given API is the one used to launch the current job.
        """
        raise NotImplementedError()

    @abstractmethod
    def rank(self) -> int:
        """
        Property containing the rank of the process.
        """
        raise NotImplementedError()

    @abstractmethod
    def local_rank(self) -> int:
        """
        Property containing the local rank of the process.
        """
        raise NotImplementedError()

    @abstractmethod
    def world_size(self) -> int:
        """
        Property containing the number of processes launched.
        """
        raise NotImplementedError()

    @abstractmethod
    def local_world_size(self) -> int:
        """
        Property containing the number of processes launched of each node.
        """
        raise NotImplementedError()

    @abstractmethod
    def num_nodes(self) -> int:
        """
        Property containing the number of nodes.
        """
        raise NotImplementedError()

    @abstractmethod
    def cpus(self) -> int:
        """
        Property containing the number of CPUs allocated to each process.
        """
        raise NotImplementedError()

    @abstractmethod
    def gpus(self) -> List[str]:
        """
        Property containing all GPUs ids.
        """
        raise NotImplementedError()

    @abstractmethod
    def nodelist(self) -> Union[str, List[str]]:
        """
        Property containing the list of nodes.
        """
        raise NotImplementedError()

    @abstractmethod
    def master_address(self) -> str:
        """
        Property containing the master node.
        """
        raise NotImplementedError()

    @abstractmethod
    def port(self) -> int:
        """
        Property containing the port to communicate with the master process.
        """
        raise NotImplementedError()

    def is_master(self) -> bool:
        """
        Detects whether the process is the master (i.e. the rank 0).
        """
        return self.rank() == 0

    def device(self) -> "torch.device":
        import torch

        if torch.cuda.is_available():
            return torch.device(f"cuda:{self.local_rank()}")
        else:
            return torch.device("cpu")

    @keep_as_func
    def init_process_group(
        self, *args, force_init: bool = False, **kwargs
    ) -> "torch.device":
        r"""
        See https://pytorch.org/docs/stable/distributed.html#torch.distributed.init_process_group
        for more infomation. Also returns the device.
        """
        import torch.distributed as dist

        _kwargs = dict(rank=self.rank(), world_size=self.world_size())
        _kwargs.update(**kwargs)

        if dist.is_initialized():
            if force_init:
                warnings.warn(
                    message=(
                        "A distributed environment had already been initialized, "
                        "but you requested to force the initialization. Attempting "
                        "to destroy the process group before recreating it."
                    ),
                    category=IdrTorchWarning,
                    stacklevel=4,
                )
                dist.destroy_process_group()
                dist.init_process_group(*args, **_kwargs)
            else:
                warnings.warn(
                    message="A distributed environment had already been initialized. Moving on.",
                    category=IdrTorchWarning,
                    stacklevel=4,
                )
        else:
            dist.init_process_group(*args, **_kwargs)
        return self.device()



import sys

# These imports won't be available at runtime, but will help VSCode completion.
from .api import API as API
from .api import AutoMasterAddressPort as AutoMasterAddressPort
from .api import decorate_methods as decorate_methods
from .api import modifiers as modifiers
from .config import *
from .interface import Interface, __version__
from .utils import IdrTorchWarning as IdrTorchWarning

sys.modules[__name__] = Interface()  # type: ignore[assignment]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Callable, Type

from ...config import __all__ as all_API_methods
from .. import API


def decorate_methods(cls: Type[API], func_to_apply: Callable) -> Type[API]:
    for obj_name in dir(cls):
        if obj_name in all_API_methods:
            decorated = func_to_apply(getattr(cls, obj_name))
            setattr(cls, obj_name, decorated)

    return cls



from .auto_master_addr_port import AutoMasterAddressPort
from .decorate_methods import decorate_methods
from .undistributed_warning import UndistributedWarning

__all__ = [
    "decorate_methods",
    "AutoMasterAddressPort",
    "UndistributedWarning",
]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
from functools import wraps
from typing import Type

from .. import API
from .decorate_methods import decorate_methods

env_variables_set: bool = False


def set_master_addr_port_env_variables(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        global env_variables_set
        if not env_variables_set:
            env_variables_set = True  # must be done before actually setting the variable to prevent stackoverflow
            os.environ["MASTER_ADDR"] = self.master_address()
            os.environ["MASTER_PORT"] = str(self.port())
            os.environ["RANK"] = str(self.rank())
            os.environ["LOCAL_RANK"] = str(self.local_rank())
            os.environ["WORLD_SIZE"] = str(self.world_size())
            os.environ["LOCAL_WORLD_SIZE"] = str(self.local_world_size())
        return func(self, *args, **kwargs)

    return wrapper


def AutoMasterAddressPort(cls: Type[API]) -> Type[API]:
    return decorate_methods(cls, func_to_apply=set_master_addr_port_env_variables)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import warnings
from typing import List

from ..utils import IdrTorchWarning
from .base import API


class TorchElasticAPI(API):
    priority: int = 9000
    name: str = "TorchElastic"

    def is_launcher(self) -> bool:
        return "TORCHELASTIC_RUN_ID" in os.environ

    def rank(self) -> int:
        return int(os.environ["RANK"])

    def local_rank(self) -> int:
        return int(os.environ["LOCAL_RANK"])

    def world_size(self) -> int:
        return int(os.environ["WORLD_SIZE"])

    def local_world_size(self) -> int:
        return int(os.environ["LOCAL_WORLD_SIZE"])

    def num_nodes(self) -> int:
        return self.world_size() // self.local_world_size()

    def cpus(self) -> int:
        return len(os.sched_getaffinity(0)) // self.local_world_size()

    def gpus(self) -> List[str]:
        return [str(i) for i in range(self.local_world_size())]

    def nodelist(self) -> List[str]:
        warnings.warn(
            "TorchElastic does not provide the whole list of nodes "
            "involved in a distributed computation. So you will only "
            "get the address of the master.",
            category=IdrTorchWarning,
            stacklevel=4,
        )
        return [self.master_address()]

    def master_address(self) -> str:
        return os.environ["MASTER_ADDR"]

    def port(self) -> int:
        return int(os.environ["MASTER_PORT"])



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
from typing import List, Union

from .base import API
from .modifiers import AutoMasterAddressPort


@AutoMasterAddressPort
class SlurmAPI(API):
    priority: int = 10000
    name: str = "Slurm"

    def is_launcher(self) -> bool:
        return "SLURM_STEP_ID" in os.environ

    def rank(self) -> int:
        return int(os.environ["SLURM_PROCID"])

    def local_rank(self) -> int:
        return int(os.environ["SLURM_LOCALID"])

    def world_size(self) -> int:
        return int(os.environ["SLURM_STEP_NUM_TASKS"])

    def local_world_size(self) -> int:
        lws = os.environ["SLURM_STEP_TASKS_PER_NODE"]
        lws = lws.split("(")[0]
        return int(lws)

    def num_nodes(self) -> int:
        return int(os.environ["SLURM_STEP_NUM_NODES"])

    def cpus(self) -> int:
        cpu = int(os.environ.get("SLURM_CPUS_PER_TASK", 0))
        return cpu or len(os.sched_getaffinity(0))

    def gpus(self) -> List[str]:
        step_gpus = os.environ.get("SLURM_STEP_GPUS", None)
        if step_gpus is not None:
            return step_gpus.split(",")
        return []

    def nodelist(self) -> Union[List[str], str]:
        compact_nodelist = os.environ["SLURM_STEP_NODELIST"]
        try:
            from hostlist import expand_hostlist
        except ImportError:
            return compact_nodelist
        else:
            return expand_hostlist(compact_nodelist)

    @staticmethod
    def get_first_host(hostlist: str) -> str:
        """
        Get the first host from SLURM's nodelist.
        Example: Nodelist="Node[1-5],Node7" -> First node: "Node1"
        Args:
            hostlist(str): the compact nodelist as given by SLURM
        Returns:
            (str): the first node to host the master process
        """
        from re import findall, split, sub

        regex = "\[([^[\]]*)\]"
        all_replacement: list[str] = findall(regex, hostlist)
        new_values = [split("-|,", element)[0] for element in all_replacement]
        for i in range(len(new_values)):
            hostlist = sub(regex, new_values[i], hostlist, count=1)
        return hostlist.split(",")[0]

    def master_address(self) -> str:
        nodelist = self.nodelist()
        if isinstance(nodelist, list):
            return nodelist[0]
        return self.get_first_host(nodelist)

    def jobid(self) -> int:
        return int(os.environ["SLURM_JOB_ID"])

    def port(self) -> int:
        return 10000 + self.jobid() % 20000



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from .api.base import API

# Endpoints redirection
rank = API.rank
local_rank = API.local_rank
world_size = API.world_size
local_world_size = API.local_world_size
num_nodes = API.num_nodes
cpus = API.cpus
gpu_ids = API.gpus
nodelist = API.nodelist
master_addr = API.master_address
master_port = API.port
is_master = API.is_master
device = API.device
init_process_group = API.init_process_group

# Aliases
ntasks = world_size
size = world_size
local_size = local_world_size
ntasks_per_node = local_world_size
nnodes = num_nodes
cpus_per_task = cpus
master_address = master_addr
init_pg = init_process_group
init = init_process_group


__all__ = [
    "rank",
    "local_rank",
    "world_size",
    "local_world_size",
    "num_nodes",
    "cpus",
    "gpu_ids",
    "nodelist",
    "master_addr",
    "master_port",
    "ntasks",
    "size",
    "local_size",
    "ntasks_per_node",
    "nnodes",
    "cpus_per_task",
    "master_address",
    "is_master",
    "device",
    "init_process_group",
    "init_pg",
    "init",
]



#! /usr/bin/env python

from setuptools import find_packages, setup

PACKAGE_NAME = "idr_torch"
VERSIONFILE = "VERSION.txt"
AUTHOR = "IDRIS"
AUTHOR_EMAIL = "assist@idris.fr"
URL = "https://www.idris.fr"

with open(VERSIONFILE, "r") as file:
    VERSION = file.read().strip()

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages(),
    include_package_data=True,
    license="MIT",
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import socket
from contextlib import closing
from typing import List, Optional

from .base import API
from .modifiers import AutoMasterAddressPort, UndistributedWarning


@UndistributedWarning
@AutoMasterAddressPort
class DefaultAPI(API):
    priority: int = 0
    name: str = "Sequential"

    def __init__(self):
        self.base_port: int = 13689
        self.current_port: Optional[int] = None

    @staticmethod
    def find_available_port() -> int:
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
            sock.bind(("localhost", 0))
            port = sock.getsockname()[1]
        return port

    def is_launcher(self) -> bool:
        return True

    def rank(self) -> int:
        return 0

    def local_rank(self) -> int:
        return 0

    def world_size(self) -> int:
        return 1

    def local_world_size(self) -> int:
        return 1

    def num_nodes(self) -> int:
        return 1

    def cpus(self) -> int:
        return len(os.sched_getaffinity(0))

    def gpus(self) -> List[str]:
        return []

    def nodelist(self) -> List[str]:
        return ["localhost"]

    def master_address(self) -> str:
        return "localhost"

    def port(self) -> int:
        if self.current_port is None:
            self.current_port = self.find_available_port()
        return self.current_port



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import socket
import warnings
from collections.abc import Iterable
from importlib.metadata import version
from functools import wraps
from inspect import isclass
from pathlib import Path
from typing import Any, List, Union

from . import __name__, __path__
from .api import API, AutoMasterAddressPort, DefaultAPI, decorate_methods
from .utils import IdrTorchWarning, warning_filter

__version__ = version(__name__)


class EmptyClass(object):
    pass


class Interface(object):
    def __init__(self):
        self._available_APIs: List[API] = []
        self.crawl_shipped_APIs()
        self.add_other_object_for_easy_access()
        self.add_API_functions()
        self.make_dir()

    @classmethod
    def add_attribute(cls, name, attribute) -> None:
        setattr(cls, name, attribute)

    def add_API_functions(self) -> None:
        from . import config

        for method_name in config.__all__:
            base_function = getattr(config, method_name)
            new_function = self.make_new_function(
                base_function.__name__,
                as_property=not getattr(base_function, "__keep_as_func__", False),
            )
            self.add_attribute(method_name, new_function)

    def make_dir(self):
        from . import config

        self.__dir: List[str] = []
        self.__dir += dir(EmptyClass())
        self.__dir += config.__all__
        self.__dir += self.__all__
        self.__dir += [
            "__version__",
        ]

    def crawl_shipped_APIs(self) -> None:
        from . import api

        self.crawl_module_for_APIs(api)

    def add_other_object_for_easy_access(self) -> None:
        from . import api
        from .api import modifiers

        self.api = api
        self.API = API
        self.AutoMasterAddressPort = AutoMasterAddressPort
        self.decorate_methods = decorate_methods
        self.IdrTorchWarning = IdrTorchWarning
        self.modifiers = modifiers
        self.__file__ = str(Path(__file__).parent / "__init__.py")
        self.__path__ = __path__
        self.__name__ = __name__
        self.__version__ = __version__
        self.__all__ = [
            "api",
            "API",
            "AutoMasterAddressPort",
            "decorate_methods",
            "IdrTorchWarning",
            "modifiers",
            "register_API",
            "get_launcher_API",
            "current_API",
            "all_APIs",
            "crawl_module_for_APIs",
            "summary",
        ]

    def __repr__(self) -> str:
        return f"<module '{self.__name__}' from '{self.__file__}'"

    def __dir__(self) -> Iterable[str]:
        return self.__dir

    def make_new_function(
        self, dest_name: str, /, as_property: bool = True
    ) -> Union[property, callable]:
        @wraps(getattr(API, dest_name))
        def redirect(self: Interface, *args, **kwargs) -> Any:
            with warnings.catch_warnings(record=True) as warning_list:
                api = self.get_launcher_API()
                output = getattr(api, dest_name)(*args, **kwargs)
            if warning_list:
                warning_filter.warn(warning_list)
            return output

        if as_property:
            return property(redirect)
        else:
            return redirect

    def register_API(self, new_API: API) -> None:
        for i, api in enumerate(self._available_APIs):
            if api.priority > new_API.priority:
                continue
            else:
                self._available_APIs.insert(i, new_API)
                break
        else:
            self._available_APIs.append(new_API)

    def get_launcher_API(self) -> API:
        for api in self._available_APIs:
            if api.is_launcher():
                return api
        return DefaultAPI()

    @property
    def current_API(self) -> str:
        return self.get_launcher_API().name

    @property
    def all_APIs(self) -> List[API]:
        return self._available_APIs

    def crawl_module_for_APIs(self, module) -> None:
        for obj_name in dir(module):
            obj = getattr(module, obj_name)
            if isclass(obj) and issubclass(obj, API) and obj is not API:
                # obj is the class so we instanciate it
                self.register_API(obj())
            elif isinstance(obj, API) and obj.__class__ is not API:
                # obj is already the instance
                self.register_API(obj)

    def summary_str(self, /, tab_length: int = 4) -> str:
        string = f"{self.current_API}(\n"
        values = {
            "rank": self.rank,
            "local_rank": self.local_rank,
            "world_size": self.world_size,
            "local_world_size": self.local_world_size,
            "cpus_per_task": self.cpus,
            "nodelist": self.nodelist,
            "hostname": socket.gethostname(),
            "master_address": self.master_addr,
            "master_port": self.master_port,
        }
        for key, value in values.items():
            string += " " * tab_length + f"{key}={value},\n"
        string += ")"
        return string

    def summary(self, /, tab_length: int = 4) -> str:
        with warnings.catch_warnings(action="ignore", category=IdrTorchWarning):
            print(self.summary_str(tab_length=tab_length))
